from re import findall, match
from typing import Any

from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.state import State
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.exceptions.model_importing_error import ModelImportingError


class EncodedFileImporter:

    @staticmethod
    def import_generated_automata(path: str = None):
        raise NotImplementedError
        # if path is None:
        #     path = "target_models"

        # for fileName in iglob(path + "/**/*.ef", recursive=True):
        #     file = open(fileName)
        #     fileLines = file.readlines()
        #     file.close()

        #     alphabet = EncodedFileImporter._get_alphabet_from_data(fileLines)
        #     del fileLines[:3]
        #     statesByName, initialStates, currentIndex = EncodedFileImporter._get_states_from_data(
        #         fileLines)
        #     del fileLines[:currentIndex + 3]
        #     EncodedFileImporter._get_transitions_from_data(
        #         fileLines, alphabet, statesByName)

        #     modelNameResults = search(r"(?:\\.*\\([^\\]+) -)", fileName)
        #     modelName = modelNameResults.group(
        #         1) if modelNameResults is not None else fileName
        #     yield DeterministicFiniteAutomaton(alphabet, initialStates,
        #                                        set(statesByName.values()), modelName)

    @staticmethod
    def _get_alphabet_from_data(fileLines: list) -> Alphabet:
        if not match(r"^(?:Alphabet: )\[(?:'([^']+)')(?:, (?:'([^']+)')+)*\]$", fileLines[0]):
            raise ModelImportingError()
        return Alphabet(frozenset(SymbolStr(symbol)
                                  for symbol in findall(r"(?:'([^']+)')", fileLines[0])))

    @staticmethod
    def _get_states_from_data(fileLines: list) -> tuple[dict[Any, State], frozenset[State], int]:
        statesByName = {}
        initialStates = set()
        currentIndex = 0

        while not fileLines[currentIndex].isspace():
            currentLine = fileLines[currentIndex]
            stateData = match(
                r"^\t(?:\"([^\"]+)\")( \(Initial\))?( \(Final\))?\n$", currentLine)

            if not stateData:
                raise ModelImportingError()

            name = stateData.group(1)
            isInitial = stateData.group(2) is not None
            isFinal = stateData.group(3) is not None

            state = State(name, isFinal)
            statesByName[name] = state
            if isInitial:
                initialStates.add(state)
            currentIndex += 1
        return statesByName, frozenset(initialStates), currentIndex

    @staticmethod
    def _get_transitions_from_data(fileLines: list, alphabet: Alphabet, statesByName: dict):
        currentIndex = 0
        while True:
            originName = match(
                r"^(?:\"([^\"]+)\"):\n$", fileLines[currentIndex])
            # TODO custom exception, hablar con franz
            assert originName is not None
            origin = statesByName[originName.group(1)]

            currentIndex += 1

            while True:
                if currentIndex >= len(fileLines):
                    return

                currentLine = fileLines[currentIndex]
                transitionData = match(
                    r"^\t(?:'([^']+)')(?: -> )\[(?:\"([^\"]+)\")(?:, (?:\"([^\"]+)\"))*\]$", currentLine)
                if not transitionData:
                    break
                symbol = alphabet[transitionData.group(1)]
                destinations = set(statesByName[name]
                                   for name in findall(r"(?:\"([^\"]+)\")", currentLine))
                origin.transitions[symbol] = destinations
                currentIndex += 1



from .python_automata_exception import PythonAutomataException

class ModelImportingError(PythonAutomataException):
    def __init__(self):
        self.message = "There was an error when importing model"



from abc import ABC, abstractmethod

from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.symbol import Symbol


class ProbabilisticModel(ABC):

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def terminal_symbol(self) -> Symbol:
        raise NotImplementedError

    @property
    @abstractmethod
    def alphabet(self) -> Alphabet:
        raise NotImplementedError

    @abstractmethod
    def sequence_probability(self, sequence: Sequence) -> float:
        raise NotImplementedError

    @abstractmethod
    def log_sequence_probability(self, sequence: Sequence) -> float:
        raise NotImplementedError

    @abstractmethod
    def last_token_probability(self, sequence: Sequence) -> float:
        raise NotImplementedError

    def last_token_probabilities_batch(self, sequences: list[Sequence], required_suffixes: list[Sequence]) -> \
            list[list[float]]:
        probas_list = list()
        for seq in sequences:
            probas = self.last_token_probabilities(seq, required_suffixes)
            probas_list.append(probas)
        return probas_list

    def last_token_probabilities(self, sequence: Sequence, required_suffixes: list[Sequence]) -> list[float]:
        probas = list()
        for suffix in required_suffixes:
            new_seq = sequence + suffix
            prob = self.last_token_probability(new_seq)
            probas.append(prob)
        return probas




from pythautomata.abstract.model_exporting_strategy import ModelExportingStrategy
from pythautomata.model_exporters.dot_exporting_strategy import DotExportingStrategy


class ImageExportingStrategy(ModelExportingStrategy):

    def export(self, model, path=None):
        graph = DotExportingStrategy().create_graph(model)
        path = self.get_path_for(path, model)
        graph.render(path, cleanup=True)



from pythautomata.base_types.alphabet import Alphabet
from pythautomata.automata.deterministic_finite_automaton import DeterministicFiniteAutomaton as DFA
import pythautomata.utilities.simple_dfa_generator as simple_dfa_generator
import math


def generate_dfa(alphabet: Alphabet, nominal_size: int, seed: int = 42) -> DFA:
    """
    Function returning a randomly generated DFA in the style of Abbadingo One Competition, generated as described in: 
        Kevin J. Lang, Barak A. Pearlmutter, and Rodney A. Price. 1998. 
        Results of the Abbadingo One DFA Learning Competition and a New Evidence-Driven State Merging Algorithm. 
        In Proceedings of the 4th International Colloquium on Grammatical Inference (ICGI '98). Springer-Verlag, Berlin, Heidelberg, 1–12.
    
    Args:
        alphabet (Alphabet): DFA alphabet. It must have size exactly 2.
        nominal_size (int): Target nominal size of the generated DFA.
        seed (int, optional): Random seed. Defaults to 42.

    Returns:
        DFA: Random DFA with distribution of sizes centered near nominal_size and depth exactly ceil(2*log(nominal_size,2)-2).
    """
    # Alphabet Size should be 2 to get the desired nominal size
    # For bigger alphabets size use nicaud_dfa_generator
    assert(len(alphabet)==2) 

    number_of_states = math.ceil((5*nominal_size)/4)
    dfa = simple_dfa_generator.generate_dfa(alphabet, number_of_states, seed)
    dfa_depth = _compute_depth(dfa)
    seed = seed + 1
    while dfa_depth != math.ceil(2*math.log(nominal_size,2)-2):
        dfa = simple_dfa_generator.generate_dfa(alphabet,  number_of_states, seed)
        dfa_depth = _compute_depth(dfa)     
        seed = seed + 1       
    return dfa
 
def _compute_depth(dfa):
    all_states = dfa.states
    initial_state = dfa.initial_state
    to_visit = [(initial_state,0)]
    depths = {}
    while len(to_visit) > 0:            
        state, depth = to_visit[-1]
        to_visit.pop()
        depths[state] = depth           
        for symbol in dfa.alphabet.symbols:
            next_states = state.next_states_for(symbol)
            for next_state in next_states:
                if next_state not in [i[0] for i in to_visit] and next_state not in depths:
                    to_visit.append((next_state, depth+1))
    
    assert(len(depths)==len(all_states))
    max_depth = max(depths.values())
    return max_depth



from unittest import TestCase
from pythautomata.utilities.simple_dfa_generator import generate_dfa
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.symbol import SymbolStr

class TestSimpleDFAGenerator(TestCase):

    def test_generated_correctly_1(self):
        binaryAlphabet = Alphabet(
            frozenset([SymbolStr('0'), SymbolStr('1')]))
        generated_automata = generate_dfa(binaryAlphabet, 80)
        self._assert_correctness(generated_automata)

    def test_generated_correctly_2(self):
        binaryAlphabet = Alphabet(
            frozenset([SymbolStr('0'), SymbolStr('1')]))
        generated_automata = generate_dfa(binaryAlphabet, 49)
        self._assert_correctness(generated_automata)

    def test_generated_correctly_3(self):
        abcdAlphabet = Alphabet(frozenset(
            [SymbolStr('a'), SymbolStr('b'), SymbolStr('c'), SymbolStr('d')]))
        generated_automata = generate_dfa(abcdAlphabet, 49)
        self._assert_correctness(generated_automata)

    def test_generated_correctly_4(self):
        alphabet012 = Alphabet(
            frozenset([SymbolStr('0'), SymbolStr('1'), SymbolStr('2')]))
        generated_automata = generate_dfa(alphabet012, 1)
        self._assert_correctness(generated_automata)

    def _assert_correctness(self, automaton):        
        self.assertTrue(self._all_states_are_rechable(automaton))

    def _all_states_are_rechable(self, automaton):
        unrechable = automaton.states.copy()
        for state in unrechable.copy():
            for destinations in state.transitions.values():
                unrechable = unrechable - destinations
        return len(unrechable) <= 1  # Hole

    def _has_final_state(self, automaton):
        return any(state.is_final for state in automaton.states)



from unittest import TestCase

from pythautomata.automata_definitions.other_automata import OtherAutomata
from pythautomata.utilities.automata_converter import AutomataConverter


class TestAutomataConvertor(TestCase):

    def test_conversion_1(self):
        automaton = OtherAutomata.get_nfa_1()
        automata_converted_to_dfa = AutomataConverter.convert_nfa_to_dfa(
            automaton)
        expected_converted_dfa = OtherAutomata.get_dfa_1()
        self.assertEqual(automata_converted_to_dfa, expected_converted_dfa)

    def test_conversion_2(self):
        automaton = OtherAutomata.get_nfa_2()
        automata_converted_to_dfa = AutomataConverter.convert_nfa_to_dfa(
            automaton)
        expected_converted_dfa = OtherAutomata.get_dfa_2()
        self.assertEqual(automata_converted_to_dfa, expected_converted_dfa)

    def test_conversion_3(self):
        automaton = OtherAutomata.get_nfa_3()
        automata_converted_to_dfa = AutomataConverter.convert_nfa_to_dfa(
            automaton)
        expected_converted_dfa = OtherAutomata.get_dfa_3()
        self.assertEqual(automata_converted_to_dfa, expected_converted_dfa)

    def test_conversion_4(self):
        automaton = OtherAutomata.get_nfa_4()
        automata_converted_to_dfa = AutomataConverter.convert_nfa_to_dfa(
            automaton)
        expected_converted_dfa = OtherAutomata.get_dfa_4()
        self.assertEqual(automata_converted_to_dfa, expected_converted_dfa)

    def test_conversion_5(self):
        automaton = OtherAutomata.get_nfa_5()
        automata_converted_to_dfa = AutomataConverter.convert_nfa_to_dfa(
            automaton)
        expected_converted_dfa = OtherAutomata.get_dfa_5()
        self.assertEqual(automata_converted_to_dfa, expected_converted_dfa)



class WeightedTransition:

    def __init__(self, next_state, weight: float):
        self.next_state = next_state
        self.weight = weight

    def __eq__(self, other):
        return isinstance(other, WeightedTransition) and self.next_state == other.next_state \
               and self.weight == other.weight

    def __hash__(self):
        return hash(self.next_state)



import pickle
from unittest import TestCase

from pythautomata.automata_definitions.tomitas_grammars import TomitasGrammars
from pythautomata.model_exporters.dot_exporting_strategy import DotExportingStrategy
from pythautomata.model_exporters.standard_dot_exporting_strategy import StandardDotExportingStrategy
from pythautomata.model_exporters.image_exporting_strategy import ImageExportingStrategy
from pythautomata.model_exporters.image_exporting_strategy_without_hole_state import ImageExportingStrategyWithoutHoleState


class TestDFAExporting(TestCase):

    def test_1(self):
        dfa = TomitasGrammars.get_automaton_1()
        name = dfa.name
        DotExportingStrategy().export(dfa, "./output_models/tests")
        ImageExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_standard"
        StandardDotExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_no_hole"
        ImageExportingStrategyWithoutHoleState().export(dfa, "./output_models/tests")

    def test_2(self):
        dfa = TomitasGrammars.get_automaton_2()
        name = dfa.name
        DotExportingStrategy().export(dfa, "./output_models/tests")
        ImageExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_standard"
        StandardDotExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_no_hole"
        ImageExportingStrategyWithoutHoleState().export(dfa, "./output_models/tests")

    def test_3(self):
        dfa = TomitasGrammars.get_automaton_3()
        name = dfa.name
        DotExportingStrategy().export(dfa, "./output_models/tests")
        ImageExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_standard"
        StandardDotExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_no_hole"
        ImageExportingStrategyWithoutHoleState().export(dfa, "./output_models/tests")

    def test_4(self):
        dfa = TomitasGrammars.get_automaton_4()
        name = dfa.name
        DotExportingStrategy().export(dfa, "./output_models/tests")
        ImageExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_standard"
        StandardDotExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_no_hole"
        ImageExportingStrategyWithoutHoleState().export(dfa, "./output_models/tests")

    def test_5(self):
        dfa = TomitasGrammars.get_automaton_5()
        name = dfa.name
        DotExportingStrategy().export(dfa, "./output_models/tests")
        ImageExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_standard"
        StandardDotExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_no_hole"
        ImageExportingStrategyWithoutHoleState().export(dfa, "./output_models/tests")

    def test_6(self):
        dfa = TomitasGrammars.get_automaton_6()
        name = dfa.name
        DotExportingStrategy().export(dfa, "./output_models/tests")
        ImageExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_standard"
        StandardDotExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_no_hole"
        ImageExportingStrategyWithoutHoleState().export(dfa, "./output_models/tests")

    def test_7(self):
        dfa = TomitasGrammars.get_automaton_7()
        name = dfa.name
        DotExportingStrategy().export(dfa, "./output_models/tests")
        ImageExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_standard"
        StandardDotExportingStrategy().export(dfa, "./output_models/tests")
        dfa.name = name + "_no_hole"
        ImageExportingStrategyWithoutHoleState().export(dfa, "./output_models/tests")



from pythautomata.base_types.state import State
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.automata.deterministic_finite_automaton import DeterministicFiniteAutomaton as DFA
import random
from pythautomata.model_exporters.image_exporting_strategy import ImageExportingStrategy
from pythautomata.model_comparators.dfa_comparison_strategy import DFAComparisonStrategy as AutomataComparator


def generate_dfa(alphabet: Alphabet, number_of_states: int = 200, seed=None, exporting_strategies=[ImageExportingStrategy()]) -> DFA:
    """
    Function returning a randomly generated DFA, with random transitions to other states and equal probability of being a final state or not.

    Args:
        alphabet (Alphabet): DFA alphabet.
        number_of_states (int, optional): Number of states the generated DFA.. Defaults to 200.
        exporting_strategies (list, optional): ExportingStrategy for the generated DFA. Defaults to [ImageExportingStrategy()].

    Returns:
        DFA: Random DFA with reachable states and exporting strategies = exporting_strategies.
        The number of states is centered around a value that depends on the alphabet size, more info in:
        Nicaud, Cyril. (2014). Random Deterministic Automata. 5-23. 10.1007/978-3-662-44522-8_2. 
    """
    if seed is not None:
        random.seed(seed)
    states = _generate_states(number_of_states)
    _add_dfa_transitions_to_states(states, alphabet.symbols)
    initial_state = next(iter(states))
    states = _remove_unreachable_states(initial_state, alphabet.symbols)
    comparator = AutomataComparator()
    return DFA(alphabet, initial_state, states, comparator=comparator, exportingStrategies=exporting_strategies)


def _generate_states(number_of_states):
    states = []
    for index in range(number_of_states):
        is_final = bool(random.getrandbits(1))
        generated_state = State(str(index), is_final)
        states.append(generated_state)
    return states


def _add_dfa_transitions_to_states(states, symbols):
    for state in states:
        for symbol in symbols:
            random_state = random.choice(states)
            state.add_transition(symbol, random_state)


def _remove_unreachable_states(initial_state, symbols):
    reachable_states = _get_reachable_states_from(initial_state, symbols)
    return reachable_states


def _get_reachable_states_from(initial_state, symbols):
    states_to_visit = [initial_state]
    visited_states = []
    while len(states_to_visit) > 0:
        state = states_to_visit.pop()
        visited_states.append(state)
        for symbol in symbols:
            next_states = state.next_states_for(symbol)
            for next_state in next_states:
                if next_state not in visited_states:
                    states_to_visit.append(next_state)
    return set(visited_states)



from pythautomata.base_types.state import State
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.automata.deterministic_finite_automaton import \
    DeterministicFiniteAutomaton
from pythautomata.model_comparators.hopcroft_karp_comparison_strategy import HopcroftKarpComparisonStrategy as HopcroftKarpComparisonStrategy

binaryAlphabet = Alphabet(frozenset((SymbolStr('0'), SymbolStr('1'))))
zero = binaryAlphabet['0']
one = binaryAlphabet['1']


class TomitasGrammarsMods:
    """
    Class containing modified versions of automata from paper:    
        Tomita, M. (1982). 
        Dynamic Construction of Finite Automata from examples using Hill-climbing. 
        Proceedings of the Fourth Annual Conference of the Cognitive Science Society 
        (p./pp. 105--108), Ann Arbor, Michigan.

    Methods
    -------   
    get_mod_automaton_5: DeterministicFiniteAutomaton
        returns a modified version of the automaton 5 from the paper
    """

    @staticmethod
    def get_all_automata():
        return [TomitasGrammarsMods.get_mod_automaton_5()]

    @staticmethod
    def get_mod_automaton_5():
        """
        method with specification of a modified version of the automaton 5 from the paper

        Recognizes strings that have an even ammount of 01 (and not taking into account 10). 
        It's a sublanguage of the fifth Tomita's grammar.

        Returns
        -------
        DeterministicFiniteAutomaton
            modified version of automaton 5 from the paper
        """
        stateQ0 = State("State 0", True)
        stateQ1 = State("State 1")
        stateQ2 = State("State 2")
        stateQ3 = State("State 3")
        stateQ0.add_transition(one, stateQ1)
        stateQ1.add_transition(zero, stateQ2)
        stateQ2.add_transition(one, stateQ3)
        stateQ3.add_transition(zero, stateQ0)

        comparator = HopcroftKarpComparisonStrategy()
        return DeterministicFiniteAutomaton(binaryAlphabet, stateQ0,
                               set([stateQ1, stateQ2, stateQ3, stateQ0]), comparator, "Tomita's grammar 5 sub language automaton")



from pythautomata.base_types.guard import Guard
from pythautomata.base_types.symbol import Symbol


class ClosedIntervalGuard(Guard):
    def __init__(self, left, right):
        self.left, self.right = left, right

    def matches(self, s: Symbol) -> bool:
        return s >= self.left and s <= self.right

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "[" + str(self.left) + " , " + str(self.right) + "]"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pythautomata',
 'pythautomata.abstract',
 'pythautomata.automata',
 'pythautomata.automata.wheighted_automaton_definition',
 'pythautomata.automata_definitions',
 'pythautomata.base_types',
 'pythautomata.boolean_algebra_learner',
 'pythautomata.exceptions',
 'pythautomata.guards',
 'pythautomata.model_comparators',
 'pythautomata.model_exporters',
 'pythautomata.regular_expressions',
 'pythautomata.tests',
 'pythautomata.utilities']

package_data = \
{'': ['*']}

install_requires = \
['graphviz', 'numpy', 'scipy>=1.7.3,<2.0.0', 'sklearn>=0.0,<0.1']

setup_kwargs = {
    'name': 'pythautomata',
    'version': '0.23.0',
    'description': "ORT's implementation of various kinds of automata",
    'long_description': '# Pythautomata\n\nPythautomata is a Python library for modeling finite state systems.\n\n## **About**\nPythautomata is developed at the Department of Artificial Intelligence and Big Data of the Universidad ORT Uruguay. Its main goal is to provide implementations for the structures needed for working in the Model Extraction Framework.\n\nModels present in the framework are:\n\n- DFA\n- NFA\n- WFA/PDFA\n- SFA\n\nAll of these can be exported in different manners, like pickle or to visual representations using Graphviz. Besides the structure representations a number of algorithms of interest are implemented, to name a few:\n\n- FSA minimization\n- FSA comparison using Hopcroft Karp equivalence\n- FSA intersection (and other boolean operations)\n\n\n## **Installation**\n\n```\npip install pythautomata\n```\n\n## **Documentation**\n\n- [**API Documentation:**](https://neuralchecker.github.io/pythautomata/index.html)\n\n\n## **Maintainers**\n\nFederico Vilensky\n\nFranz Mayr\n\nFederico Pan\n\n## Colaborators\n\n\n',
    'author': 'Federico Vilensky',
    'author_email': 'fedevilensky@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



from pythautomata.base_types.symbol import Symbol
from pythautomata.base_types.guard import Guard

class EqualityGuard(Guard):
    def __init__(self, s:Symbol):
        self._s = s

    def matches(self, s: Symbol) -> bool:
        return self._s == s

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return str(self._s)


import unittest
from pythautomata.utilities import dfa_operations
from pythautomata.automata_definitions.other_automata import OtherAutomata

class TestDFAOperations(unittest.TestCase):

    def test_join_1(self):
        automatonA = OtherAutomata.get_uneven_number_of_as_automaton()
        automatonB = OtherAutomata.get_uneven_number_of_symbols_automaton()
        target_join = OtherAutomata.get_uneven_number_of_as_and_symbols_automaton()
        joinedAutomaton = dfa_operations.join_DFAs(automatonA, automatonB)
        self.assertEqual(target_join, joinedAutomaton)

    def test_join_2(self):
        automatonA = OtherAutomata.get_empty_automaton()
        automatonB = OtherAutomata.get_sigma_star_automaton()
        target_join = OtherAutomata.get_empty_automaton()
        joinedAutomaton = dfa_operations.join_DFAs(automatonA, automatonB)
        self.assertEqual(target_join, joinedAutomaton)    

    def test_union_1(self):
        automatonA = OtherAutomata.get_uneven_number_of_as_automaton()
        automatonB = OtherAutomata.get_uneven_number_of_symbols_automaton()
        target_union = OtherAutomata.get_uneven_number_of_as_or_symbols_automaton()
        unionAutomaton = dfa_operations.union_DFAs(automatonA, automatonB)
        self.assertEqual(target_union, unionAutomaton)  

    def test_union_2(self):
        automatonA = OtherAutomata.get_empty_automaton()
        automatonB = OtherAutomata.get_sigma_star_automaton()
        target_join = OtherAutomata.get_sigma_star_automaton()
        unionAutomaton = dfa_operations.union_DFAs(automatonA, automatonB)
        self.assertEqual(target_join, unionAutomaton)
